@@ -7,7 +7,7 @@ rm $CONDA_PREFIX/bin/pumaX_testing
 
 echo "Testing PuMA GUI"
 pumaGUI &
-killall pumaGUI
+pkill pumaGUI
 
 echo "Testing pumapy"
 cd python/test
