@@ -1,7 +1,7 @@
 from skimage.io import imread
 import numpy as np
 import pickle
-from pumapy import Workspace
+from pumapy.utilities.workspace import Workspace
 from os import path
 from glob import glob
 from vtkmodules.util.numpy_support import vtk_to_numpy
@@ -9,22 +9,31 @@ from vtkmodules.vtkIOXML import vtkXMLImageDataReader, vtkXMLUnstructuredGridRea
 from vtkmodules.vtkIOLegacy import vtkDataSetReader
 
 
+def io_logs(ws, filename, input=True):
+    if input:
+        s = ["Importing", "from "]
+    else:
+        s = ["Exporting", "to "]
+    ws.log.log_section(s[0] + " Domain " + s[1] + filename)
+    ws.log.write_log()
+
+
 def import_3Dtiff(filename, voxel_length=1e-6, import_ws=True):
     """ Function to io 3D tiff
 
-    :param filename: filepath and name
-    :type filename: string
-    :param voxel_length: size of a voxel side
-    :type voxel_length: float, optional
-    :param import_ws: if True returns a puma.Workspace, otherwise a ndarray
-    :type import_ws: bool, optional
-    :return: domain
-    :rtype: Workspace or ndarray
+        :param filename: filepath and name
+        :type filename: string
+        :param voxel_length: size of a voxel side
+        :type voxel_length: float
+        :param import_ws: if True returns a puma.Workspace, otherwise a ndarray
+        :type import_ws: bool
+        :return: domain
+        :rtype: pumapy.Workspace or np.ndarray
 
-    :Example:
-    >>> import pumapy as puma
-    >>> ws_tiff = puma.import_3Dtiff(puma.path_to_example_file("50_artfibers.tif"), 1.3e-6, import_ws=True)
-    >>> ws_tiff.get_shape()
+        :Example:
+        >>> import pumapy as puma
+        >>> ws_tiff = puma.import_3Dtiff(puma.path_to_example_file("50_artfibers.tif"), 1.3e-6, import_ws=True)
+        >>> ws_tiff.get_shape()
     """
     print("Importing " + filename + " ... ", end='')
 
@@ -42,21 +51,22 @@ def import_3Dtiff(filename, voxel_length=1e-6, import_ws=True):
     if import_ws:
         ws = Workspace.from_array(nparray)
         ws.set_voxel_length(voxel_length)
+        io_logs(ws, filename)
         return ws
     return nparray
 
 
 def import_bin(filename):
-    """ Import a puma.Workspace to binary (.pumapy extension)
+    """ Import a pumapy.Workspace to binary (.pumapy extension)
 
-    :param filename: filepath and name
-    :type filename: string
-    :return: True if successful, False otherwise.
-    :rtype: bool
+        :param filename: filepath and name
+        :type filename: string
+        :return: True if successful, False otherwise.
+        :rtype: bool
 
-    :Example:
-    >>> import pumapy as puma
-    >>> ws_binary = puma.import_bin(puma.path_to_example_file("fibers_with_orientation.pumapy"))
+        :Example:
+        >>> import pumapy as puma
+        >>> ws_binary = puma.import_bin(puma.path_to_example_file("fibers_with_orientation.pumapy"))
     """
     print("Importing " + filename + " ... ", end='')
 
@@ -67,25 +77,26 @@ def import_bin(filename):
     ws = pickle.load(pumapy_file)
     pumapy_file.close()
     print("Done")
+    io_logs(ws, filename)
     return ws
 
 
 def import_vti(filename, voxel_length=None, import_ws=True):
     """ Function to import either legacy VTK file (.vtk) or vtkImageData (.vti)
 
-    :param filename: filepath and name
-    :type filename: string
-    :param voxel_length: voxel_length. If None, voxel_length from the vtk file is used
-    :type voxel_length: float, optional
-    :param import_ws: True returns a puma.Workspace, otherwise a list of ndarrays
-    :type import_ws: bool, optional
-    :return: if import_ws is True, then it returns a Workspace.
-        if import_ws is False, it returns a dictionary of ndarrays as {"name1": data1, "name2": data2 ...}
-    :rtype: Workspace or dict(string: ndarray)
+        :param filename: filepath and name
+        :type filename: string
+        :param voxel_length: voxel_length. If None, voxel_length from the vtk file is used
+        :type voxel_length: float
+        :param import_ws: True returns a puma.Workspace, otherwise a list of ndarrays
+        :type import_ws: bool
+        :return: if import_ws is True, then it returns a Workspace.
+            if import_ws is False, it returns a dictionary of ndarrays as {"name1": data1, "name2": data2 ...}
+        :rtype: pumapy.Workspace or {str: np.ndarray}
 
-    :Example:
-    >>> import pumapy as puma
-    >>> ws_vtk = puma.import_vti(puma.path_to_example_file("fibers_with_orientation.vti"))
+        :Example:
+        >>> import pumapy as puma
+        >>> ws_vtk = puma.import_vti(puma.path_to_example_file("fibers_with_orientation.vti"))
     """
     print("Importing " + filename + " ... ", end='')
 
@@ -161,6 +172,7 @@ def import_vti(filename, voxel_length=None, import_ws=True):
             ws.set_voxel_length(vtkobject.GetSpacing()[0])
         else:
             ws.set_voxel_length(voxel_length)
+        io_logs(ws, filename)
         return ws
 
     else:
@@ -183,13 +195,15 @@ def import_vti(filename, voxel_length=None, import_ws=True):
         return nparray_list
 
 
-def import_weave_vtu(filename):
+def import_weave_vtu(filename, from_texgen_gui=False):
     """ Import TexGen vtu weave in a Workspace
 
-    :param filename: file path and name
-    :type filename: string
-    :return: voxelized weave from TexGen
-    :rtype: Workspace
+        :param filename: file path and name
+        :type filename: string
+        :param from_texgen_gui: voxel grid exported from the TexGen GUI (Windows) or from TexGen inside PuMA
+        :type from_texgen_gui: bool
+        :return: voxelized weave from TexGen
+        :rtype: pumapy.Workspace
     """
 
     if not path.exists(filename):
@@ -206,32 +220,29 @@ def import_weave_vtu(filename):
     vtkobject = reader.GetOutputDataObject(0)
 
     dims = path.split(filename[:-4])[1].split('_')[-3:]
-
-    # ORIGINAL TEXGEN
-    # Number Of Arrays: 6
-    # Array 0 name = YarnIndex  <-- transferring to ws
-    # Array 1 name = YarnTangent
-    # Array 2 name = Location
-    # Array 3 name = VolumeFraction
-    # Array 4 name = SurfaceDistance
-    # Array 5 name = Orientation  <-- transferring to ws
-    # yarn_index = vtk_to_numpy(vtkobject.GetCellData().GetArray(0)) + 1
-    # orientation = vtk_to_numpy(vtkobject.GetCellData().GetArray(5))
-    # yarn_index = yarn_index.reshape(int(dims[0]), int(dims[1]), int(dims[2]), order="F")
-    # orientation = orientation.reshape(int(dims[0]), int(dims[1]), int(dims[2]), 3, order="F")
-
-    # MODIFIED TEXGEN
-    # Number Of Arrays: 6
-    # Array 0 name = YarnIndex  <-- transferring to ws
-    # Array 1 name = Orientation  <-- depends on has_orientation
     yarn_index = vtk_to_numpy(vtkobject.GetCellData().GetArray(0)) + 1
-    yarn_index = yarn_index.reshape(int(dims[0]), int(dims[1]), int(dims[2]), order="F")
-    ws = Workspace.from_array(yarn_index)
+    ws = Workspace.from_array(yarn_index.reshape(int(dims[0]), int(dims[1]), int(dims[2]), order="F"))
 
-    if vtkobject.GetCellData().GetNumberOfArrays() == 2:
-        orientation = vtk_to_numpy(vtkobject.GetCellData().GetArray(1))
+    if vtkobject.GetCellData().GetNumberOfArrays() > 2:
+        if from_texgen_gui:
+            # ORIGINAL TEXGEN (GUI in Windows)
+            # Number Of Arrays: 6
+            # Array 0 name = YarnIndex  <-- transferring to ws
+            # Array 1 name = YarnTangent (unnecessary)
+            # Array 2 name = Location (unnecessary)
+            # Array 3 name = VolumeFraction (unnecessary)
+            # Array 4 name = SurfaceDistance (unnecessary)
+            # Array 5 name = Orientation  <-- transferring to ws
+            orientation = vtk_to_numpy(vtkobject.GetCellData().GetArray(5))
+        else:
+            # MODIFIED TEXGEN wrapped in PuMA
+            # Number Of Arrays: 6
+            # Array 0 name = YarnIndex  <-- transferring to ws
+            # Array 1 name = Orientation  <-- depends on export_orientation in export_weave_vtu
+            orientation = vtk_to_numpy(vtkobject.GetCellData().GetArray(1))
         orientation = orientation.reshape(int(dims[0]), int(dims[1]), int(dims[2]), 3, order="F")
         ws.set_orientation(orientation)
 
     print("Done")
+    io_logs(ws, filename)
     return ws
